##### Configuration

log ?= $(if $(test),,1)
profile ?=
reverify ?= 1
exts ?= csv tsv txt xml
test_n ?= 2
tablesSort ?= plots organisms stems specimens

##### Vars/functions

selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))

# Paths
datasrc := $(notdir $(realpath .))
root := $(selfDir_uZPPqC)..
bin := $(root)/bin
mappings := $(root)/mappings

# Make
SHELL := /bin/bash
selfMake = $(MAKE) --makefile=../input.Makefile
subMake = $(MAKE) $(@:$(root)/%=%) --directory=$(root)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# OS
os := $(shell uname)
forOs = $(patsubst %,%-$(filter Linux Darwin,$(os)),$(1))

# Formatting
SED = sed -$(if $(filter Darwin,$(os)),E,r)

# System
date = $(shell date +"%Y-%m-%d-%H-%M-%S")

# Terminal
termCols := $(shell tput cols)
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
MKDIR = mkdir -p
mkdir = $(MKDIR) $(@D)
CP = cp -p
diff = diff --unified=2
diffVerbose = $(if $(verbose),diff --side-by-side --left-column\
--width=$(termCols),$(diff))

# Commands
selfMap = $(bin)/cols 0 0
psqlOpts := --set ON_ERROR_STOP=1 --quiet
psqlAsBien := $(bin)/psql_vegbien $(psqlOpts)

# SVN
addDir = $(if $(wildcard $(1)/),svn add --depth=empty $(1),svn mkdir $(1))
setSvnIgnore = svn propset svn:ignore $(2) $(1)
define addDirWithIgnore
$(addDir)
$(setSvnIgnore)
endef

##### General targets

all: _always maps ;

.SUFFIXES: # turn off built-in suffix rules
.SECONDARY: # don't automatically delete intermediate files
.DELETE_ON_ERROR: # delete target if recipe fails

_always:
.PHONY: _always

clean: _always
	$(RM) $(all)

remake: _always clean
	+$(selfMake)
# re-run make so that cache of existing files is reset

make_script = ./$< >$@

%/: % _always ;

# Must come before `%: %.make` to override it
src/%: src/%.make _always
	(set -x; $(make_script)) 2>>$<.log
.PRECIOUS: src/% # save partial outputs of aborted src make scripts

%: %.make _always
	$(make_script)

##### SVN

add: _always
	$(call setSvnIgnore,.,$$'')
	$(call addDirWithIgnore,src,$$'*')
	$(call addDirWithIgnore,maps,$$'.~*')
	$(call addDirWithIgnore,import,$$'*')
	$(call addDirWithIgnore,test,$$'*.out\n*.xml')
	$(call addDirWithIgnore,verify,$$'*.out')

##### Installation

reinstall: _always uninstall install ;

##### Existing maps discovery

srcMap := maps/src.%.csv
fullViaMap := maps/%.full.csv
directMap := maps/VegBIEN.%.csv
allViaMaps := $(filter-out $(srcMap) $(fullViaMap) $(directMap),\
$(wildcard maps/*.csv))
via := $(firstword $(sort $(basename $(basename $(notdir $(allViaMaps))))))

coreMap := $(mappings)/$(via)-VegBIEN.%.csv
coreSelfMap := $(mappings)/$(via).self.%.csv

viaMaps := $(wildcard $(tablesSort:%=maps/$(via).%.csv))
viaMaps += $(filter-out $(viaMaps) $(fullViaMap),$(wildcard maps/$(via).*.csv))

autogenMaps := $(subst $(via).,VegBIEN.,$(viaMaps))
directMaps := $(autogenMaps) $(filter-out $(autogenMaps),\
$(wildcard maps/VegBIEN.*.csv))
tables := $(directMaps:maps/VegBIEN.%.csv=%)

##### Sources

srcs = $(shell $(bin)/sort_filenames $(wildcard $(exts:%=src/*.$*.%)))
isXml = $(filter %.xml,$(firstword $(srcs)))
catSrcs = $(bin)/cat$(if $(isXml),,_csv) $(srcs)
withCatSrcs = $(catSrcs:$(bin)/%=$(bin)/with_%) --

# Usage: `make inputs/self/cat|tail +4` (removes echoed make commands)
cat: $(addprefix cat-,$(tables)) _always ;

cat-%: _always
	$(catSrcs)

##### Maps building

# Maps to (try to) build are added to one of these
maps :=
createOnlyMaps :=

srcMaps := $(wildcard maps/src.*.csv)

mkSrcMap = env datasrc=$(datasrc) $(withCatSrcs) $(bin)/src_map >$@

# Autogen src maps with known table names
maps/src.%.csv: # no prereqs so only built if doesn't exist
	$(if $(srcs),$(if $(isXml),,$(mkSrcMap)))
# only build if CSV srcs exist for that table name
# Try all table names
createOnlyMaps += $(tablesSort:%=maps/src.%.csv)

# Must come before $(root)/% to override it
$(coreSelfMap): _always
	-+$(subMake)
# ignore errors if $(coreSelfMap) does not exist

# Via maps cleanup
ifneq ($(filter maps/.%.last_cleanup,$(MAKECMDGOALS)),)
maps/.$(via).%.csv.last_cleanup: maps/$(via).%.csv $(coreMap)
	$(bin)/in_place $< env ignore=1 $(bin)/subtract $(word 2,$+) 0 1
	touch $@
else
$(viaMaps): _always
	$(selfMake) $(@:maps/%=maps/.%.last_cleanup)
endif

joinSrcMap = $(if $(wildcard maps/src.$*.csv),$(bin)/in_place $@\
$(bin)/intersect maps/src.$*.csv 0)

makeFullCsv = $(if $(shell test -e $(word 2,$+) && echo t),\
env ignore=1 $(bin)/union <$+|,<$< )$(bin)/sort_map >$@
# can't use $(wildcard) because it won't recheck file after $(coreSelfMap) runs

maps/$(via).%.full.csv: maps/$(via).%.csv $(coreSelfMap)
	$(makeFullCsv)
	$(joinSrcMap)
maps += $(patsubst maps/%.csv,maps/%.full.csv,$(viaMaps))

maps/VegBIEN.%.csv: maps/$(via).%.full.csv $(coreMap)
	$(bin)/join <$+|$(bin)/sort_map >$@
maps += $(autogenMaps)

maps: $(createOnlyMaps) $(maps) _always ;

all += $(maps)

##### Maps validation

missing_mappings: _always missing_join_mappings missing_input_mappings ;

missing_%_mappings: _always # stem is one of join|input
	@echo $(emph)"Missing $* mappings:"$(endEmph)
	@+$(selfMake) remake 2>&1\
|$(SED) -n 's/^.*No.* $* mapping for ([0-9A-Za-z_-]+).*$$/\1/p'\
$(if $(filter join,$*),|$(bin)/ucase_first 0)|sort|uniq

##### External dependencies

$(root)/%: _always
	+$(subMake)
.PRECIOUS: $(root)/% # let ext. dir's Makefile decide whether to delete on error

##### Mapping

dbExport := $(firstword $(wildcard src/db.*.sql))
inputFiles := $(wildcard $(exts:%=src/*.%))

+maps = $(filter maps/% $(mappings)/%,$(+_))
<in = $(firstword $(filter-out $(+maps),$(+_)))
map = $(if $(<in),<$(<in),\
$(if $(srcs),$(withCatSrcs),\
$(if $(mapEnv),env $(mapEnv),\
$(error No input file src/*.$*.{$(exts)}))))\
$(root)/map $(+maps)
map2db = env out_database=vegbien $(map)

##### Import to VegBIEN

ifneq ($(dbExport)$(inputFiles),)

profileTest = $(if $(profile),$(if $(test),1))
profileOnly = -env profile_to=/dev/fd/3 $(map2db) 3>&1 1>&2|\
$(bin)/profile_stats /dev/fd/0

log_ = import/$*$(if $(n),.n=$(n),).$(date).log
trace = $(log_:.log=.trace)
import = -$(if $(profileTest),$(profileOnly),(set -x; "time" env commit=1\
$(if $(profile),profile_to=$(trace)) $(map2db)) $(if $(log),\
$(if $(n),,>>$(log_))) 2>&1$(if $(log),$(if $(n),|tee -a $(log_))))
# don't abort on import errors, which often relate to invalid input data

import: $(addprefix import-,$(tables)) _always ;

import-%: maps/VegBIEN.%.csv _always
	$(import)
# default:
import-%: _always ;

else
import: _always ;
endif

##### Log files from import

logs := $(wildcard import/*.log import/*.trace)

rm_logs: _always
	$(RM) $(logs)

##### Verification of import

verify: $(addprefix verify-,$(tables)) _always ;

verify-%: verify/%.ref verify/%.out _always
	-$(diffVerbose) $(+_)
# don't abort on verification errors, which are expected during development
# default:
verify-%: verify/%.out _always
	$(if $(shell test -e $< && echo t),cat $<)
# don't run if verify/%.out's default do-nothing action was used
# can't use $(wildcard) because it won't recheck file after verify/%.out is run

verify = $(if $(reverify),"time" $(psqlAsBien) --set=datasource="'$(datasrc)'"\
--no-align --field-separator=$$'\t' --pset=footer=off --pset=null=NULL <$< >$@)

verify/%.out: $(mappings)/verify.%.sql _always
	$(verify)
# default:
verify/%.out: _always ;

all += $(wildcard verify/*.out)

ifneq ($(dbExport),)
%.ref: %.ref.sql
	$(dbAsBien) $(db) <$< >$@
endif

##### Editing import

import/rotate: _always
	echo "UPDATE party SET organizationname = organizationname||'.$(date)'\
WHERE organizationname = '$(datasrc)';"|$(psqlAsBien)

import/rm: _always
	echo "DELETE FROM party WHERE organizationname = '$(datasrc)';"|\
$(psqlAsBien)

##### Testing

hasOwnRef = $(filter-out %.2-step.xml,$@)
testRef = $(1:.2-step.xml=.xml).ref

define runTest
@echo "Testing $(abspath $@)..."
>$@ env test=1 n=$(test_n) $(1)
@(set -x; $(diff) $(call testRef,$@) $@) 2>&1 || { e=$$?;\
$(if $(wildcard $(call testRef,$@)),,cat $@;)\
$(if $(hasOwnRef),\
echo $(emph)"To accept new test output:"$(endEmph);\
echo "$(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile";\
,\
echo $(emph)"Note: The preceding failed test is compared to another test's\
output"$(endEmph);\
echo $(emph)"When it fails, this always indicates a bug"$(endEmph);\
)\
exit $$e;}
endef

test2File = $(call runTest,$(map))

tests :=

test/$(via).%.xml: maps/$(via).%.full.csv _always
	$(test2File)
tests += test/$(via).%.xml

test/VegBIEN.%.xml: maps/VegBIEN.%.csv _always
	$(test2File)
tests += test/VegBIEN.%.xml

test/VegBIEN.%.2-step.xml: test/$(via).%.xml $(coreMap) _always
	-$(test2File)
# Don't abort tester if only 2-step test fails, as it's often finicky
tests += test/VegBIEN.%.2-step.xml

test/import.%.out: maps/VegBIEN.%.csv _always
	$(call runTest,$(map2db))
tests += test/import.%.out

testOutputs := $(foreach test,$(tests),$(tables:%=$(test)))

.PRECIOUS: $(testOutputs) # save outputs of failed tests so they can be accepted

test: _always $(testOutputs) ;

all += $(testOutputs)

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	$(CP) $* $(call testRef,$*)

##### Input-type-specific

# Each input type needs var $(mapEnv) and targets install, uninstall

#### DB export

ifneq ($(dbExport),)

dbEngineExt := $(subst .,,$(suffix $(basename $(notdir $(dbExport)))))
db := $(datasrc)

### Installation

install: _always db ;

uninstall: _always rm_db ;

### DB-engine-specific

# Each DB engine needs vars $(dbEngine), $(dbAsBien) and targets db, rm_db

ifeq ($(dbEngineExt),my)

dbEngine := MySQL

bienPassword := $(shell cat $(root)/config/bien_password)
mysqlAs = mysql --user=$(1) --password='$(bienPassword)'
mysqlAsRoot := $(call mysqlAs,root)
dbAsBien := $(call mysqlAs,bien)

dbExists = $(shell echo "SHOW DATABASES LIKE '$(db)';"|$(mysqlAsRoot))

define createDb
echo "CREATE DATABASE $(db) DEFAULT CHARACTER SET latin1;"|$(mysqlAsRoot)
-$(mysqlAsRoot) --database=$(db) <$<
endef
# ignore errors in db import so that GRANT will still be run

db: $(dbExport) _always
	$(if $(dbExists),,$(createDb))
	echo "GRANT SELECT ON $(db).* TO 'bien'@'localhost';"|$(mysqlAsRoot)

rm_db: _always
	-echo "REVOKE ALL ON $(db).* FROM 'bien'@'localhost';"|$(mysqlAsRoot)
	echo "DROP DATABASE IF EXISTS $(db);"|$(mysqlAsRoot)
# ignore errors if grant not defined

## Unrecognized DB engine

else
$(error The DB filename $(dbExport) must be db.my.sql)
endif

### Other input types

else

install: _always ;
uninstall: _always ;

endif

#### DB connection info

ifneq ($(dbEngine),)
# Must come after dbEngine is set
mapEnv := in_engine=$(dbEngine) in_database=$(db)
endif
